// Copyright (C) 2019-2020 Terence Kelly.  All rights reserved.
// Hand-over-hand locking example program.  Intended to accompany
// article in USENIX _;login:_ magazine circa late 2020.
// Compile as C99 or C11; for gcc, compile & link with "-pthread".

#include <assert.h>
#include <errno.h>
#include <pthread.h>
#include <stdio.h>
#include <unistd.h>

typedef struct node { pthread_mutex_t m;
                      int data;
                      struct node *next; } node_t;

#define            PMI PTHREAD_MUTEX_INITIALIZER
static node_t E = {PMI, 5, NULL},
              D = {PMI, 4, &E},
              C = {PMI, 3, &D},
              B = {PMI, 2, &C},
              A = {PMI, 1, &B},
           head = {PMI, 0, &A};  // dummy node

#define NAME(p)  (   &head == (p) ? "head"                  \
                   : &A    == (p) ? "A"                     \
                   : &B    == (p) ? "B"                     \
                   : &C    == (p) ? "C"                     \
                   : &D    == (p) ? "D"                     \
                   : &E    == (p) ? "E"                     \
                   : NULL  == (p) ? "NULL" : (assert(0), "?") )

#define DIE(s)      ( perror(s), assert(0), 1 )
#define PT(f, ...)  ( ( errno = pthread_ ## f (__VA_ARGS__)) \
                       && DIE(#f)                              )

#define LOCK(p)         PT(mutex_lock,   (&((p)->m)))
#define UNLK(p)  ((void)PT(mutex_unlock, (&((p)->m))), (p)=NULL)

static pthread_mutex_t   pm = PMI;  // print mutex
#define printf(...)                                \
  do { PT(mutex_lock,   &pm); printf(__VA_ARGS__); \
       PT(mutex_unlock, &pm); } while (0)

static void * hoh(void * ID) {
  char *id =     (char *)ID;
  node_t *p, *n;  // "previous" follows "next" down the list
  printf("%s: begin\n", id);
  for (p = &head, LOCK(p); NULL != (n = p->next); p = n) {
    // A:  *p locked & might be dummy head
    //     *n not yet locked & can't be head
    LOCK(n);
    // B:  we may remove *n here
    UNLK(p);
    // C:  best place to inspect *n or insert node after *n
    printf("%s: node %s @ %p data %d\n",
           id, NAME(n), (void *)n, n->data);
    n->data++;
  }
  // D
  sleep(1) && DIE("sleep");  // stall for "convoy" interleaving
  UNLK(p);
  printf("%s: end\n", id);
  return id;
}

#define NTHREADS 4
int main(void) {
  pthread_t t[NTHREADS];  int i;  void *tr;
  char m1[] = "1st (serial) traversal",
       m2[] = "2nd (serial) traversal",
       id[NTHREADS][3] = {{"T0"},{"T1"},{"T2"},{"T3"}};
  hoh((void *)m1);
  hoh((void *)m2);
  printf("\nmain: going multi-threaded:\n\n");
  for (i = 0; i < NTHREADS; i++)
    PT(create, &t[i], NULL, hoh, (void *)id[i]);
  for (i = 0; i < NTHREADS; i++) {
    PT(join, t[i], &tr);
    printf("main: joined %s\n", (char *)tr);
  }
  printf("\nmain: all threads finished\n");
  return 0;
}
